from typing import ClassVar, Dict, Optional, cast

from .interface import OutletInterface


class SNMPOutlet(OutletInterface):
    type: ClassVar[str] = "snmp"

    def __init__(
        self,
        *,
        host: str,
        query_oid: str,
        query_on_value: object,
        query_off_value: object,
        update_oid: str,
        update_on_value: object,
        update_off_value: object,
        read_community: str = "public",
        write_community: str = "private",
    ) -> None:
        self.host = host
        self.query_oid = query_oid
        self.query_on_value = query_on_value
        self.query_off_value = query_off_value
        self.update_oid = update_oid
        self.update_on_value = update_on_value
        self.update_off_value = update_off_value
        self.read_community = read_community
        self.write_community = write_community

        # Import this here to pay less startup time cost.
        import pysnmp.hlapi as snmplib  # type: ignore
        import pysnmp.proto.rfc1902 as rfc1902  # type: ignore
        self.snmplib = snmplib
        self.rfc1902 = rfc1902

        if type(query_on_value) != type(query_off_value):
            raise Exception("Unexpected differing types for query on and off values!")
        if type(update_on_value) != type(update_off_value):
            raise Exception("Unexpected differing types for update on and off values!")

    def serialize(self) -> Dict[str, object]:
        return {
            'host': self.host,
            'query_oid': self.query_oid,
            'query_on_value': self.query_on_value,
            'query_off_value': self.query_off_value,
            'update_oid': self.update_oid,
            'update_on_value': self.update_on_value,
            'update_off_value': self.update_off_value,
            'read_community': self.read_community,
            'write_community': self.write_community,
        }

    @staticmethod
    def deserialize(vals: Dict[str, object]) -> OutletInterface:
        return SNMPOutlet(
            host=cast(str, vals['host']),
            query_oid=cast(str, vals['query_oid']),
            query_on_value=vals['query_on_value'],
            query_off_value=vals['query_off_value'],
            update_oid=cast(str, vals['update_oid']),
            update_on_value=vals['update_on_value'],
            update_off_value=vals['update_off_value'],
            read_community=cast(str, vals['read_community']),
            write_community=cast(str, vals['write_community']),
        )

    def query(self, value: object) -> Optional[object]:
        if isinstance(self.query_on_value, int):
            try:
                return int(str(value))
            except ValueError:
                return None
        raise NotImplementedError(f"Type of query value {type(self.query_on_value)} not supported!")

    def update(self, value: bool) -> object:
        if isinstance(self.update_on_value, int):
            return self.rfc1902.Integer(self.update_on_value if value else self.update_off_value)
        raise NotImplementedError(f"Type of update value {type(self.update_on_value)} not supported!")

    def getState(self) -> Optional[bool]:
        iterator = self.snmplib.getCmd(
            self.snmplib.SnmpEngine(),
            self.snmplib.CommunityData(self.read_community, mpModel=0),
            self.snmplib.UdpTransportTarget((self.host, 161), timeout=1.0, retries=0),
            self.snmplib.ContextData(),
            self.snmplib.ObjectType(self.snmplib.ObjectIdentity(self.query_oid)),
        )

        for response in iterator:
            errorIndication, errorStatus, errorIndex, varBinds = response
            if errorIndication:
                return None
            elif errorStatus:
                return None
            else:
                for varBind in varBinds:
                    actual = self.query(varBind[1])
                    if actual == self.query_on_value:
                        return True
                    if actual == self.query_off_value:
                        return False
                    return None
        return None

    def setState(self, state: bool) -> None:
        iterator = self.snmplib.setCmd(
            self.snmplib.SnmpEngine(),
            self.snmplib.CommunityData(self.write_community, mpModel=0),
            self.snmplib.UdpTransportTarget((self.host, 161)),
            self.snmplib.ContextData(),
            self.snmplib.ObjectType(self.snmplib.ObjectIdentity(self.update_oid), self.update(state)),
        )
        next(iterator)



import os
from setuptools import setup  # type: ignore

dir_path = os.path.dirname(os.path.realpath(__file__))
with open(os.path.join(dir_path, "README.md"), "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name='smartoutlet',
    version='0.2.1',
    description='Collection of utilities for interfacing with various PDUs and smart outlets.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='DragonMinded',
    author_email='dragonminded@dragonminded.com',
    license='Public Domain',
    url='https://github.com/DragonMinded/smartoutlet',
    package_data={"smartoutlet": ["py.typed"]},
    packages=[
        'smartoutlet',
    ],
    install_requires=[
        'requests',
        'pysnmplib',
    ],
    python_requires=">=3.6",
)



import argparse
import inspect
import os
import sys

from smartoutlet import ALL_OUTLET_CLASSES


def addargs(parser: argparse.ArgumentParser) -> None:
    logloc = os.path.abspath(os.path.join(os.getcwd(), "daemon.log"))
    parser.add_argument(
        "--daemon",
        action="store_true",
        help="Use a local daemon to speed up fetch and set requests.",
    )
    parser.add_argument(
        "--port",
        metavar="PORT",
        type=int,
        help="the port which the daemon will listen on, defaults to 54545",
    )
    parser.add_argument(
        "--log",
        metavar="LOG",
        type=str,
        default=logloc,
        help=f"the logfile we will write daemon logs to, defaults to {logloc}",
    )


def cli(mode: str) -> int:
    outlettypes = ', '.join(c.type for c in ALL_OUTLET_CLASSES)

    if mode == "fetch":
        parser = argparse.ArgumentParser(description="Fetch the state of a smart outlet or PDU.", add_help=False)
    else:
        parser = argparse.ArgumentParser(description="Set the state of a smart outlet or PDU.", add_help=False)
    parser.add_argument(
        "type",
        metavar="TYPE",
        type=str,
        nargs="?",
        default="",
        help=f"the type of outlet you are controlling, valid values are {outlettypes}",
    )
    addargs(parser)

    knownargs, _ = parser.parse_known_args()

    # Rebuild parser with help enabled so we can get actual help strings.
    if mode == "fetch":
        parser = argparse.ArgumentParser(description="Fetch the state of a smart outlet or PDU.")
    else:
        parser = argparse.ArgumentParser(description="Set the state of a smart outlet or PDU.")
    parser.add_argument(
        "type",
        metavar="TYPE",
        type=str,
        help=f"the type of outlet you are controlling, valid values are {outlettypes}",
    )
    addargs(parser)

    # Add outlet-specific arguments.
    for clz in ALL_OUTLET_CLASSES:
        if clz.type.lower() == knownargs.type.lower():
            # Figure out arguments to add for this outlet.
            signature = inspect.signature(clz.__init__)
            for param in signature.parameters.values():
                if param.name == "self":
                    continue
                if param.default is inspect.Parameter.empty:
                    parser.add_argument(
                        param.name,
                        metavar=param.name.upper(),
                        type=param.annotation,
                        help=f"{param.annotation.__name__} parameter",
                    )
                else:
                    parser.add_argument(
                        f"--{param.name}",
                        type=param.annotation,
                        default=param.default,
                        help=f"{param.annotation.__name__} parameter, defaults to {param.default}",
                    )
            break
    else:
        if knownargs.type:
            print(f"Unrecognized outlet type {knownargs.type}!", os.linesep, file=sys.stderr)
            parser.print_help()
            return 1

    if mode == "set":
        parser.add_argument(
            "state",
            metavar="STATE",
            type=str,
            help="the state you want to set the outlet to, valid values are on and off",
        )
    args = vars(parser.parse_args())
    if mode == "set":
        state = args['state'].lower() == "on"
        del args['state']

    # Figure out arguments to add for this outlet.
    signature = inspect.signature(clz.__init__)
    constructor_args = {}
    for param in signature.parameters.values():
        if param.name == "self":
            continue
        constructor_args[param.name] = args[param.name]

    if args['daemon']:
        # Import this here so we don't pay the cost otherwise.
        from smartoutlet.daemon import OutletProxy
        inst = OutletProxy.deserialize({'type': clz.type, 'port': args['port'], 'log': args['log'], **constructor_args})
    else:
        inst = clz.deserialize(constructor_args)

    if mode == "fetch":
        state = inst.getState()
        if state is None:
            print("unknown")
        else:
            print("on" if state else "off")
    else:
        inst.setState(state)

    return 0



from abc import ABC, abstractmethod
from typing import ClassVar, Dict, Optional


class OutletInterface(ABC):
    type: ClassVar[str]

    @abstractmethod
    def serialize(self) -> Dict[str, object]:
        ...

    @staticmethod
    @abstractmethod
    def deserialize(vals: Dict[str, object]) -> "OutletInterface":
        ...

    @abstractmethod
    def getState(self) -> Optional[bool]:
        ...

    @abstractmethod
    def setState(self, state: bool) -> None:
        ...
